import cv2 as cv
import numpy as np

def grid(base, dimentions, images, scale=0.5):
	# 1. SCALE IMAGE
	base = cv.resize(base, (0, 0), fx=scale, fy=scale)
	images = [cv.resize(image, (0, 0), fx=scale, fy=scale) for image in images]
	# 2. COMPLETE DIMENTIONS IF MISSING
	for i, image in enumerate(images):
		if len(image.shape) < 3:
			images[i] = cv.cvtColor(image, cv.COLOR_GRAY2BGR)
	# 3. CREATE GRID
	missing = dimentions[0]*dimentions[1] - len(images)
	if missing < 0:
		raise Exception('Wrong grid dimentions')
	for i in range(missing):
		images.append(np.zeros(base.shape, dtype=np.uint8))
	grid = np.array(images);
	grid = grid.reshape( (dimentions[0], dimentions[1], base.shape[0], base.shape[1], base.shape[2]) )
	# 4. STACK IMAGES
	return np.vstack( [np.hstack(row[:]) for row in grid] )

def getBoxesOffset(im, boxes):
	im_w = im.shape[1]
	im_h = im.shape[0]
	offsets = []

	for box in boxes:
		x, y, w, h = box
		xc, yc = (x + int(w/2), y + int(h/2))

		x_off = 2*xc/im.shape[1] - 1
		y_off = 1 - 2*yc/im.shape[0]

		offsets.append( (x_off, y_off) )

	return offsets

def drawBoxes(im, boxes):
	for box in boxes:
		x, y, w, h = box
		xc, yc = (x + int(w/2), y + int(h/2))
		cv.rectangle( im, (x, y), (x + w, y + h), (255, 250, 255), 2 )
	return im

def drawBoxesPos(im, boxes):
	cv.line(
		im,
		(0, int(im.shape[0]/2)),
		(im.shape[1], int(im.shape[0]/2)),
		(0, 255, 0), 2)
	cv.line(
		im,
		(int(im.shape[1]/2), 0),
		(int(im.shape[1]/2), im.shape[0] ),
		(0, 255, 0), 2)

	for box in boxes:
		x, y, w, h = box
		xc, yc = (x + int(w/2), y + int(h/2))
		cv.circle( im, (xc, yc) , 1, (130, 250, 255), 2 )
		cv.line( im, (xc, yc), (int( im.shape[1]/2), yc), (130, 250, 255), 2 )
		cv.line( im, (xc, yc), (xc, int( im.shape[0]/2)), (130, 250, 255), 2 )

	return im



import cv2 as cv
import numpy as np

class Features:
	def __init__(self, im_source=None, features=500):
		if im_source is None:
			raise Exception('Must input a grayscale image')
		self.corners = np.float32([
			[0, 0], [0, im_source.shape[0]],
			[im_source.shape[1], im_source.shape[0]], [im_source.shape[1], 0]
		]).reshape(-1, 1, 2)
		self.bf = cv.BFMatcher()
		self.features = features
		self.orb = cv.ORB_create(nfeatures=features)
		self.kp1, self.des1 = self.orb.detectAndCompute(im_source, None)
		self.im_source_kp = cv.drawKeypoints(im_source.copy(), self.kp1, None)
		self.im_source = im_source

		self.kp2, self.des2 = (None, None)
		self.im_target_kp = None
		self.im_target = None

		self.im_poly = None

	def loadTarget(self, im):
		self.im_target = im.copy()

	def getMatches(self, distance):
		self.kp2, self.des2 = self.orb.detectAndCompute(self.im_target, None)
		self.im_target_kp = cv.drawKeypoints(self.im_target.copy(), self.kp2, None)

		if len(self.kp2) <= 1:
			raise Exception('Not enough keypoints in target')

		matches = self.bf.knnMatch(self.des1, self.des2, k=2) # RESEARCH

		good = []
		for m, n in matches:
			if m.distance < distance *  n.distance:
				good.append( m )
		return good

	def matchPoints(self, matches):
		return cv.drawMatchesKnn(
			self.im_source, self.kp1,
			self.im_target, self.kp2,
			[matches], None, flags=2)

	def getBoxes(self, matches, min_matches):
		self.im_poly = self.im_target.copy()
		boxes = []

		if len(matches) >= min_matches:
			src_points = np.float32(
				[self.kp1[m.queryIdx].pt for m in matches]
			).reshape(-1, 1, 2)
			dst_points = np.float32(
				[self.kp2[m.trainIdx].pt for m in matches]
			).reshape(-1, 1, 2)

			matrix, mask = cv.findHomography(src_points, dst_points, cv.RANSAC, 5) # RESEARCH

			if matrix is None:
				raise Exception('Cannot find homography in matches')

			dts = cv.perspectiveTransform(self.corners, matrix)
			cv.polylines(self.im_poly, [np.int32(dts)], True, (0, 255, 0), 3)
			boxes.append(cv.boundingRect(np.int32(dts)))

		return boxes

if __name__ == '__main__':
	import sys
	from os import path
	sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )
	import cv_tools

	if len(sys.argv) == 2:
		im_source = cv.imread(sys.argv[1])
		#im_source = cv.resize(im_source, (0, 0), fx=0.5, fy=0.5)
	elif len(sys.argv) == 1:
		raise Exception('Must input a path to an image')
		exit()
	else:
		raise Exception('Too many args')
		exit()

	my_feature = Features(im_source, 1000)

	cam = cv.VideoCapture(0)
	cam.set(cv.CAP_PROP_FRAME_WIDTH, 320)
	cam.set(cv.CAP_PROP_FRAME_HEIGHT, 240)

	while True:
		if cv.waitKey(1) & 0xFF == ord('q'):
			break

		rec, frame = cam.read()
		my_feature.loadTarget(frame)
		frame_matches = None
		boxes = []

		try:
			matches = my_feature.getMatches(0.75)
			frame_matches = my_feature.matchPoints(matches)
			boxes = my_feature.getBoxes(matches, 20)
		except Exception as e:
			print(e)
			continue

		if frame_matches is not None:
			cv.imshow('image matches', frame_matches)

		frame_boxes = cv_tools.drawBoxes(frame.copy(), boxes)
		cv_tools.drawBoxesPos(frame_boxes, boxes)
		frame_grid = cv_tools.grid(frame, (1, 3), [
			frame, my_feature.im_poly, frame_boxes
		])
		cv.imshow('grid', frame_grid)

	cam.release()
	cv.destroyAllWindows()



import cv2 as cv
import numpy as np

class Colorspace:
	def __init__(self, hsv_settings = None):
		if hsv_settings is None:
			self.lower = np.int32( [0, 0, 0] )
			self.upper = np.int32( [179, 255, 255] )
		elif type(hsv_settings) is list and np.int32(hsv_settings).shape == (2,3):
			self.lower = np.int32( [hsv_settings[0]] )
			self.upper = np.int32( [hsv_settings[1]] )
		else:
			Colorspace.loadSettings(self, hsv_settings)
		self.im_mask = None
		self.im_cut = None
		self.im_edges = None
		self.im_contours = None

	def loadSettings(self, settings):
		try:
			with open(settings, 'r') as f:
				lines = f.read().split('\n')
				self.lower = np.int32( [value for value in lines[0].split(',')] )
				self.upper = np.int32( [value for value in lines[1].split(',')] )
		except Exception as e:
			print(e)
			exit()

	def dumpSettings(self, output='last.log'):
		with open(output, 'w') as f:
			f.write('{},{},{}\n{},{},{}'.format(
				self.lower[0], self.lower[1], self.lower[2],
				self.upper[0], self.upper[1], self.upper[2])
			)

	@staticmethod
	def createSliders():
		cv.namedWindow('sliders')
		cv.createTrackbar('hmin', 'sliders',   0, 179, lambda x: x)
		cv.createTrackbar('hmax', 'sliders', 179, 179, lambda x: x)
		cv.createTrackbar('smin', 'sliders',   0, 255, lambda x: x)
		cv.createTrackbar('smax', 'sliders', 255, 255, lambda x: x)
		cv.createTrackbar('vmin', 'sliders',   0, 255, lambda x: x)
		cv.createTrackbar('vmax', 'sliders', 255, 255, lambda x: x)

	def updateHSV(self):
		self.lower = np.array([
			cv.getTrackbarPos('hmin', 'sliders'),
			cv.getTrackbarPos('smin', 'sliders'),
			cv.getTrackbarPos('vmin', 'sliders')
		])
		self.upper = np.array([
			cv.getTrackbarPos('hmax', 'sliders'),
			cv.getTrackbarPos('smax', 'sliders'),
			cv.getTrackbarPos('vmax', 'sliders')
		])

	def getMaskBoxes(self, im_base, im_hsv, min_area=20, scale=0.2):
		self.im_mask = cv.inRange(im_hsv, self.lower, self.upper)
		self.im_cut = cv.bitwise_and(im_base, im_base, mask=self.im_mask)
		self.im_edges = cv.Canny(self.im_mask, 100, 100)
		self.im_contours = im_base.copy()

		boxes_within = []
		contours, _ = cv.findContours(self.im_edges, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)	# <- RESEARCH

		for c in contours:
			area = cv.contourArea(c)
			if area >= min_area:
				perimeter = cv.arcLength( c, True )
				points = cv.approxPolyDP( c, scale * perimeter, True )
				boxes_within.append(cv.boundingRect(points))
				cv.drawContours(self.im_contours, c, -1, (255, 255, 255), 2)					# <- RESEARCH

		return boxes_within

if __name__ == '__main__':
	import sys
	from os import path
	sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )
	import cv_tools

	colorspace = Colorspace()
	sample_mode =  True

	if len(sys.argv) == 1:
		colorspace.createSliders()
	elif len(sys.argv) == 2:
		sample_mode =  False
		colorspace.loadSettings(sys.argv[1])
	else:
		raise Exception('Too many args')
		exit()

	cam = cv.VideoCapture(0)
	cam.set(cv.CAP_PROP_FRAME_WIDTH, 320)
	cam.set(cv.CAP_PROP_FRAME_HEIGHT, 240)

	while True:
		if sample_mode:
			colorspace.updateHSV()

		rec, frame = cam.read()
		if not rec:
			raise Exception('Cam is not recording')

		frame_blur = cv.GaussianBlur(frame, (9, 9), 150) 		# smoothes the noise
		frame_hsv = cv.cvtColor(frame_blur, cv.COLOR_BGR2HSV)	# convert BGR to HSV

		boxes = colorspace.getMaskBoxes(frame, frame_hsv, 150)	# get boxes
		offsets = cv_tools.getBoxesOffset(frame, boxes)			# get boxes offset from the center of the frame
		for i, offset in enumerate(offsets):
			print(i, offset)

		frame_out = cv_tools.drawBoxes(frame.copy(), boxes)
		frame_out = cv_tools.drawBoxesPos(frame_out, boxes)

		frame_grid = cv_tools.grid(frame, (2, 3),[
			frame_hsv, colorspace.im_contours, frame_out,
			colorspace.im_cut, colorspace.im_mask, colorspace.im_edges
		])

		cv.imshow('grid', frame_grid)

		if cv.waitKey(1) & 0xFF == ord('q'):
			break

	cam.release()
	cv.destroyAllWindows()

	if sample_mode:
		colorspace.dumpSettings()



from picamera.array import PiRGBArray
from picamera import PiCamera
from threading import Thread
from time import sleep

class PiCamStream:
	def __init__(self, resolution=(320, 240), fps=32, **kargs):
		self.camera = PiCamera()
		self.camera.resolution = resolution
		self.camera.brightness = 55
		self.camera.contrast = 10

		self.fps = fps
		self.rawCapture = PiRGBArray(self.camera, size=resolution)

		self.stream = self.camera.capture_continuous(
			self.rawCapture,
			format='bgr',
			use_video_port=True
		)

		self.current_frame = None
		self.stop = False

	def start(self):
		cam_thread = Thread(target=self.update, args=())
		cam_thread.daemon = True
		cam_thread.start()

	def update(self):
		for frame in self.stream:
			self.current_frame = frame.array
			self.rawCapture.truncate(0)
			if self.stop:
				self.stream.close()
				self.rawCapture.close()
				self.camera.close()
				break

	def close(self):
		self.stop = True




import setuptools

with open("README.md", "r") as fh:
	readme = fh.read()

setuptools.setup(
	name="cv-recon",
	version="0.0.4",
	author="Arturo Aguilar Lagunas",
	author_email="aguilar.lagunas.arturo@gmail.com",
	description="A computer vision toolkit focused in color detection and feature matching using OpenCV. It allows you to easily start the picamera in case you're using a Raspberry PI.",
	long_description=readme,
	long_description_content_type="text/markdown",
	url="https://github.com/AguilarLagunasArturo/cv-recon",
	packages=setuptools.find_packages(),
	keywords='opencv color-detection feature-matching computer-vision picamera',
	classifiers=[
		"Development Status :: 4 - Beta",
		"Programming Language :: Python",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
)
